import numpy as np
from abc import ABC, abstractmethod


class ImpurityMeasure(ABC):
    """
    Abstract base class for impurity measures, such as Gini and Entropy.
    """
    def __init__(self) -> None:
        self.name = self.__class__.__name__

    def __call__(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        """
        Calls the _loss method to compute the impurity measure value of the input target array.
        """
        return self._loss(targets, weights)

    @abstractmethod
    def _loss(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        """
        Computes the impurity measure value of the input target array.
        """
        pass


class Gini(ImpurityMeasure):
    """
    A class for computing the Gini impurity measure.
    """
    def __init__(self) -> None:
        super().__init__()

    def _loss(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        """
        Computes the Gini impurity measure value of the input target array.
        """
        uniTargets, counts = np.unique(targets, return_counts=True) # Get unique target classes and their counts
        weights = np.ones(uniTargets.shape[0]) if weights is None else weights # If no weights provided, consider all classes equally important
        p = weights * counts / len(targets) # Compute the probability of each class in the target array
        return (p * (1 - p)).sum() # Compute the Gini impurity measure value


class Entropy(ImpurityMeasure):
    """
    A class for computing the Entropy impurity measure.
    """
    def __init__(self) -> None:
        super().__init__()

    def _loss(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        """
        Computes the Entropy impurity measure value of the input target array.
        """
        uniTargets, counts = np.unique(targets, return_counts=True) # Get unique target classes and their counts
        weights = np.ones(uniTargets.shape[0]) if weights is None else weights # If no weights provided, consider all classes equally important
        p = weights * counts / len(targets) # Compute the probability of each class in the target array
        return - (p * np.log2(p)).sum() # Compute the Entropy


class MAE(ImpurityMeasure):
    """
    Computes the mean absolute error (MAE) of the given targets.
    """
    def __init__(self) -> None:
        super().__init__()

    def _loss(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        targetsMean = np.mean(targets)
        return np.sum(np.abs(targets - targetsMean)) / len(targets)


class MSE(ImpurityMeasure):
    """
    Computes the mean squared error (MSE) of the given targets.
    """
    def __init__(self) -> None:
        super().__init__()

    def _loss(self, targets: np.ndarray, weights: np.ndarray = None) -> float:
        targetsMean = np.mean(targets)
        return np.sum((targets - targetsMean) ** 2) / len(targets)
